#
# $Id: fawww_pgms.pl 35 2009-10-28 18:29:25Z wrp $
# $Revision: 38 $
#

# this set of lists and hashes defines most of the behavior of the
# different run modes.  There are two major functions in
# fasta_www.cgi/FASTA_WWW.pm, do_form() and do_search().
#
# The pages presented by do_form() are defined by %form_list{},
# %page_links{}, and %page_link_list;
#
# The pages produced by do_search() are defined by %run_list{} with help
# from %fa_opt_params{} and %pgm_dev{}
#

#
# in-line text for %form_list and $pgm_link_list;
#
require "./fawww_pgm_text.pl";

# %form_list{form_name} provides the information required by do_form()
# to produce an input/selection form from a template.
#
# form_list->{form_name} = {
#      tmpl => template_file,
#      pgm_ref => [list of programs to be selected from],
#      pgm_def => "default_program",
#      "inputs" sets template vars in form from input parameters
#      inputs => { arg1 => {TMP_VAR1 => value, TMP_VAR2 => \&value_func, ...},
#
#      inp_dep_list => [ array of required parameters with dependencies]
#      e.g. in mselect, the query can be derived from the msa_query
#      inp_deps => { param_name => { "alt_param_name", \&trans_func()},...}
#      outputs => { TMPL_VAR1 => value, ...}
#      every template should get a RUN_MODE
#

%form_list =
  (
###
   'misc1'=>
   {tmpl=>"misc1.tmpl",
    pgm_ref=>\@pgm_mlist,
    pgm_def=>"pkd",
    inputs => 
    { query => {SEARCH_QUERY => "this", SEARCH_FRM_ACC => "selected"}
    },
    outputs =>
    { TITLE => qq(Misc. Protein Analysis),
      RUN_MODE=>'misc1_rx',
    }
   },
###
   'shuffle' =>
   {tmpl=>"compare.tmpl", pgm_ref=>\@pgm_shuff_list,
    pgm_def=>"rss",
    inputs =>
    {query => {SEARCH_QUERY => "this", SEARCH_FRM_ACC => "selected",
	       SSR => \&get_query_range},
     query2 => {SEARCH_QUERY2 => "this", SEARCH_FRM_ACC2 => "selected"}
    },
    outputs => 
    {SUBMIT => "Shuffle Sequence",
     SSR_FLAG => '1',
     Q2_FILE_UP => 1,
     HAVE_SSR2 => '1',
     SEG_FLAG => '1',
     TITLE => qq(PRSS/PRFX Sequence Shuffling),
     OPTION1 => $shuff_opt1,
     OPTION2 => $shuff_opt2,
     MSA_PSSM_FILE => $shuff_msa_opt,
     RUN_MODE=>'shuffle_r',
    },
   },
###
   'compare' =>
   { tmpl=>"compare.tmpl",
     pgm_ref=>[@pgm_flist, @pgm_slist],
     pgm_def=>"fap",
     CAN_REMOTE => 1,
     inputs =>
    { query => {SEARCH_QUERY => "this", SEARCH_FRM_ACC => "selected",
		SSR => \&get_query_range},
      query2 => {SEARCH_QUERY2 => "this", SEARCH_FRM_ACC2 => "selected"}
    },
    outputs =>
    { SUBMIT => 'Compare Sequences',
      Q2_FILE_UP => 1,
      SSR_FLAG => '1',
      HAVE_SSR2 => '1',
      MSA_PSSM_FILE => $shuff_msa_opt,
      TITLE => qq(FASTA Sequence Comparison),
      OPTION1 => $comp_opt1,
      OPTION2 => $comp_opt2,
      OPTION3 => $comp_opt3,
      SEG_FLAG => 1,
      RUN_MODE => 'compare_r',
      REM_RUN_MODE=>'compare_r',
    },
   },
###
   'lalign' =>
   { tmpl=>"compare.tmpl", pgm_ref=>\@pgm_lalign_list,
     pgm_def=>"lal",
     inputs =>
     { query => {SEARCH_QUERY => "this", SEARCH_FRM_ACC => "selected",
		 SSR => \&get_query_range},
       query2 => {SEARCH_QUERY2 => "this", SEARCH_FRM_ACC2 => "selected",
		  SSR2 => \&get_query_range,},
     },
     outputs =>
     { SUBMIT => 'Align Sequences',
       TITLE => qq(LALIGN/PLALIGN local alignments),
       OPTION1 => $lalign_opt1,
       OPTION2 => $lalign_opt2,
       RUN_MODE => 'lalign_x',
       Q2_FILE_UP => 1,
       SSR_FLAG => '1',
       HAVE_SSR2 => '1',
       SSR_FLAG => '1',
     },
   },
###
   'blast' =>
   { tmpl=>"select.tmpl", pgm_ref=>\@pgm_blist, pgm_def=>"bp",
     lib_ref=> \@blp_list,
     CAN_REMOTE => 1,
     inputs =>
     { query => {SEARCH_QUERY => "this", SEARCH_FRM_ACC => "selected"},
       remote => {SHOW_REMOTE => "this", RUNMODE => 'remote'},
     },
     outputs =>
     { TITLE => qq(BLAST Sequence Comparison),
       RUN_MODE=>'blast_r',
       REM_RUN_MODE => 'blast_r',
       OPTION1 => $blast_opt1,
     },
   },
###
   'select' =>
   { tmpl=>"select.tmpl",
     pgm_ref=>[@pgm_flist, @pgm_slist], pgm_def=>"fap",
     lib_env=> $FAST_LIBS,
     CAN_REMOTE => 1,
     inputs =>
     { query => {SEARCH_QUERY => "this", SEARCH_FRM_ACC => "selected",
		 SSR => \&get_query_range},
       remote => {SHOW_REMOTE => "this", RUNMODE => 'remote'},
       acc => {SEARCH_FRM_ACC=>"selected"}
     },
     outputs =>
     { RUN_MODE=>'search',
       REM_RUN_MODE=>'search',
       TITLE => qq(FASTA Sequence Comparison),
       OPTION1 => $select_opt1,
       LIB_OPT => $lib_opt,
       OPTION2 => $select_opt2,
       OPTION3 => $select_opt3,
       OPTION4 => $select_opt4,
       MSA_PSSM_FILE => $shuff_msa_opt,
     },
   },
###
   'selectg' =>
   { tmpl=>"select.tmpl", 
     pgm_ref=>[@pgm_flist, @pgm_slist], pgm_def=>"fap",
     lib_env=> $FAST_GNMS,
     inputs =>
     { query => {SEARCH_QUERY => "this", SEARCH_FRM_ACC => "selected",
		 SSR => \&get_query_range},
       remote => {SHOW_REMOTE => "this"},
     },
     outputs =>
     { RUN_MODE=>'searchg', 
       REM_RUN_MODE=>'searchg', 
       TITLE => qq(FASTA Sequence Comparison),
       OPTION1 => $select_opt1,
       OPTION2 => $select_opt2,
       LIB_OPT => $lib_opt,
       OPTION3 => $select_opt3,
       MSA_PSSM_FILE => $shuff_msa_opt,
     },
   },
###
   'mselect' =>
   { tmpl=>"mselect.tmpl",
     pgm_ref=>\@pgm_pssmlist, pgm_def=>"pbp",
     lib_ref=> \@blp_list,
     inputs =>
     { query => {SEARCH_QUERY => "this", SEARCH_FRM_FA => "selected"},
       msa_query => {MSA_QUERY => "this"}
     },
     inp_dep_list => [qw(msa_query query)],
     inp_deps => {msa_query => ["msa_query", \&clean_mquery],
		  query => ["msa_query", \&query_from_mquery],
		  },
     outputs =>
     { TITLE => qq(PSSM Sequence Comparison),
       MSA_FILE_UP => 1,
       RUN_MODE=>'msearch_x',
       OPTION1 => $mselect_opt1,
       MSA_PSSM_FILE => $shuff_msa_opt,
     },
   },

   'hmm_select' =>
   { tmpl=>"hmm_select.tmpl",
     pgm_ref=>\@pgm_hmmlist, pgm_def=>"hmms",
     lib_ref=> \@hmmp_list,
     inputs =>
     { hmm_query => {HMM_QUERY => "this"}
     },
     outputs =>
     { TITLE => qq(HMM Sequence Comparison),
       HMM_FILE_UP => 1,
       RUN_MODE=>'hmm_search',
     },
   },

   'hmm_select2' =>
   { tmpl=>"hmm_select2.tmpl",
     pgm_ref=>\@pgm_hmmlist, pgm_def=>"hmms",
     lib_ref=> \@hmmp_list,
     inputs =>
     { hmm_query => {HMM_QUERY => "this"}
     },
     outputs =>
     { TITLE => qq(HMM Sequence Comparison),
       HMM_FILE_UP => 1,
       RUN_MODE=>'hmm_search',
       SUBMIT => 'Scan HMM',
       Q2_FILE_UP => 1,
       MSA_PSSM_FILE => $shuff_msa_opt,
       TITLE => qq(FASTA Sequence Comparison),
       OPTION2 => $comp_opt2,
       OPTION3 => $comp_opt3,
       RUN_MODE => 'hmm_search',
       REM_RUN_MODE=>'hmm_search',
    },
   },
  );

#
# lots of programs have gap penalties and scoring matrices
#
%fa_opt_params = 
    (
     gap => ["-f", \&get_safe_number],
     ext => ["-g", \&get_safe_number],
     smatrix => ["", \&get_smatrix],
     segflag => ["-S", \&get_option],
     );

# %run_list{run_name} provides the information required by do_search()
# to run a program, given the input query an parameters
#
# run_list->{run_name} = {
#      pgm_ref => [list of programs to be selected from],
#      n_q => number of queries (n_q = 1 for searches, 2 for comparison)
#      use_query1 => 1 (put the @ in the command line)
#      have_ssr => get ssr for query1
#      query2_type => tmp -> temporary file, q2 for \@, lib for lib selection
#      get_lib_sub => get a library from lib_p or lib_n
#      lib_env => FAST_LIBS environment
#      remote => 1 -> can remote
#      pgm_args => "default string after program before other options"
#   opts and post_opts are both parsed by the same code
#      opts => { param1 => [ "option_str1", \&get_func, default],
#		 param2 => [ "option_str2", \&get_func, default],
#	  default is the value provided if the param1 is not given
#

%run_list = (
     'misc1_rx'=> {
	 indirect => {
	     seg => misc1_seg,
	     cho => misc1_rss,
	     gor => misc1_rss,
	     pkd => misc1_pkd,
	     }
     },
     'misc1_pkd'=> {
	 pgm_ref=>\@pgm_mlist, n_q => 1, sq_type => 1, 
	 q_arg => "query",
	 qt_arg => "q_type",
	 qf_arg => "query_file",
	 use_query1 => 1, no_html => 1,
	 post_opts => {
	     kd_window => ["", \&get_safe_number]
	     },
     },
# 'misc1_seg' does not use use_query1 (no \@)
     'misc1_seg'=> {
	 pgm_ref=>\@pgm_mlist, n_q => 1, sq_type => 1, 
	 q_arg => "query",
	 qt_arg => "q_type",
	 qf_arg => "query_file",
	 no_html => 1,
	 post_opts => {
	     seg_domain => [ ["","- -q -z 1", "- -p"],
			     \&put_indexed_args,"- -q -z 1"]
	     },
     },
     'misc1_rss'=> {	# used for "gor", "cho"
	 pgm_ref=>\@pgm_mlist, n_q => 1, sq_type => 1, 
	 q_arg => "query",
	 qt_arg => "q_type",
	 qf_arg => "query_file",
	 use_query1 => 1, no_html => 1,
     },
####
     'shuffle_r' => {
	 pgm_ref=> \@pgm_shuff_list, n_q => 2,
	 pgm_args => "-q -w 80 -m 6 -Z 10000 -A -H -E 1000.0",
	 use_query1 => 1, query2_type => 'tmp',
	 have_ssr => 1, have_ssr2 => 1,
	 have_url_ref =>"", use_ktup => 1,
	 link_url_ref => "",
	 opts => 
	 { %fa_opt_params,
	   shuff_cnt => ["-k",\&get_safe_number, "-k 200"],
	   shuff_w => ["-v 20", \&get_option],
	   msa_asn_file => ["-P \"%s 2\"", \&get_file2file],
	   msa_asn_rid => ["-P \"%s 2\"", \&get_rid2file],
         },
	 post_opts => { ktup => ["", \&get_safe_number] },
	 rem_files => {
	     msa_asn_file => [\&get_file2file, 'msa_asn_file'],
	     msa_asn_rid => [\&get_rid2file, 'msa_asn_file']
	 },
	 footer => $fa_footer,
     },
####
     'compare_r' => {
	 pgm_ref=>[@pgm_flist, @pgm_slist, @pgm_hlist],
	 n_q => 2, have_ssr => 1, have_url_ref =>"",
	 pgm_args => "-q -w 80 -m 6 -z 11 -Z 10000",
	 use_query1 => 1, query2_type => 'tmp',
	 use_ktup => 1,
	 link_url_ref => "",
	 opts => { %fa_opt_params, 
		   eval => ["-E",\&get_safe_number],
		   msa_asn_file => ["-P \"%s 2\"", \&get_file2file],
		   msa_asn_rid => ["-P \"%s 2\"", \&get_rid2file],
	       },
	 post_opts => { ktup => ["", \&get_safe_number]},
	 rem_files => {
	     msa_asn_file => [\&get_file2file, 'msa_asn_file'],
	     msa_asn_rid => [\&get_rid2file, 'msa_asn_file']
	 },
	 footer => $fa_footer,
     },

     'lalign_x' => {
	 indirect => { 
            lal => lalign_r,
	    lald => lalign_r,
	    pal => plalign_r,
            pald => plalign_r,
	 }
      },

     'lalign_r' => {
	 pgm_ref=>\@pgm_lalign_list, n_q => 2,
	 pdfdev => 1, have_ssr => 1, have_ssr2 => 1,
	 use_query1 => 1, query2_type => 'tmp', 
	 pgm_args => '-q -w 80 -m 6',
	 link_url_ref => "",
	 opts => {
	     show_ident => ["-I", \&get_option],
	     gap => ["-f", \&get_safe_number],
	     ext => ["-g", \&get_safe_number],
	     smatrix => ["", \&get_smatrix],
	 },
	 footer => $fa_footer,
     },

     'plalign_r' => {
	 pgm_ref=>\@pgm_lalign_list, n_q => 2,
	 pdfdev => 1, have_ssr => 1, have_ssr2 => 1,
	 use_query1 => 1, query2_type => 'tmp', no_html => 1,
	 pgm_args => '-q -m 11',
	 opts => {
	     show_ident => ["-I", \&get_option],
	     gap => ["-f", \&get_safe_number],
	     ext => ["-g", \&get_safe_number],
	     smatrix => ["", \&get_smatrix],
	 },
	 footer => $fa_footer,
     },

     'search' => {
	 pgm_ref=>[@pgm_flist, @pgm_slist, @pgm_hlist],
	 n_q => 1, lib_env=> $FAST_LIBS, remote=>1, have_ssr=>1,
	 run_bkgd => 1,
	 pgm_args => "-q -w 80 -m 9i -m 6 -b 500 -d 250",
	 q_arg => "query",
	 qt_arg => "q_type",
	 qf_arg => "query_file",
	 use_query1 => 1, query2_type => 'lib',
	 get_lib_sub => \&get_lib,
	 use_ktup => 1,
	 link_url_ref => "",
	 opts => {
	     %fa_opt_params,
	     db_range =>["-M", \&get_safe_range],
	     eval => ["-E",\&get_safe_number],
	     etop => ["-F",\&get_safe_number],
	     aln_type => [["","-m 1","-m 2"],\&put_indexed_args,""],
	     show_hist => [["-H",""],\&put_indexed_args,"-H"],
	     zstat => ["-z",\&get_safe_number],
	     msa_asn_file => ["-P \"%s 2\"", \&get_file2file],
	     msa_asn_rid => ["-P \"%s 2\"", \&get_rid2file],
	     rem_asn_text => ["-P \"%s 2\"", \&get_text2file],
	 },
	 post_opts => { ktup => ["", \&get_safe_number]},
	 rem_files => {
	     msa_asn_file => [\&get_file2file, 'msa_asn_file'],
	     msa_asn_rid => [\&get_rid2file, 'msa_asn_file']
	 },
	 footer => $fa_footer,
     }, 

     'searchg' => {
	 pgm_ref=>[@pgm_flist, @pgm_slist, @pgm_hlist],
	 n_q => 1, lib_env=> $FAST_GNMS, remote=>1, have_ssr =>1,
	 run_bkgd => 1,
	 pgm_args => "-q -w 80 -m 6 ",
	 use_query1 => 1, query2_type => 'lib',
	 get_lib_sub => \&get_lib,
	 use_ktup => 1,
	 link_url_ref =>"g",
	 opts => {
	     %fa_opt_params,
	     db_range =>["-M", \&get_safe_range],
	     eval => ["-E",\&get_safe_number],
	     etop => ["-F",\&get_safe_number],
	     zstat => ["-z",\&get_safe_number],
	     msa_asn_file => ["-P \"%s 2\"", \&get_file2file],
	     msa_asn_rid => ["-P \"%s 2\"", \&get_rid2file],
	 },
	 post_opts => { ktup => ["", \&get_safe_number]},
	 rem_files => {
	     msa_asn_file => [\&get_file2file, 'msa_asn_file'],
	     msa_asn_rid => [\&get_rid2file, 'msa_asn_file']
	 },
	 footer => $fa_footer,
     }, 

     'blast_r' => {
	 pgm_ref=>\@pgm_blist, lib_ref => \@blp_list,
	 n_q => 1, remote=>1,
	 run_bkgd => 1,
	 pgm_args => "-T T",
	 opts => {
	     eval => ["-e", \&get_safe_number],
	     comp_stat => ["-C 0", \&get_neg_option],
	     segflag => [["",qq(-F F),qq(-F "S"), qq(-F "m S")], \&put_indexed_args,],
	     p_lib => ["-d", \&get_blib],
	     gap => ["-G", \&get_safe_number],
	     ext => ["-E", \&get_safe_number],
	     smatrix => ["", \&get_bmatrix],
	 },
	 footer => $bl_footer,
     },

     'msearch_x' => {
	 indirect => { 
	     pbp => msearch_pbp,
	     psw => msearch_psw,
	     pgg => msearch_psw,
	     pgl => msearch_psw,
	 }
     },

     'msearch_pbp' => {
	 pgm_ref=>\@pgm_pssmlist, lib_ref => \@blp_list,
	 run_bkgd => 1,
	 n_q => 1,
	 pgm_args => "-T T",
	 opts => {
	     msa_query => ["-B %s", \&get_text2file ],
	     iter => ["-j", \&get_safe_number, "-j 3"],
	     eval => ["-h", \&get_safe_number, "-h 0.001"],
	     comp_stat => ["-t F", \&get_neg_option],
	     p_lib => ["-d", \&get_blib],
	     msa_asn_file => [qq(-q 2 -R %s),\&get_file2file],
	     msa_asn_rid => [qq(-q 2 -R %s), \&get_rid2file],
	 },
	 footer => $bl_footer,
     },

     'msearch_psw' => {
	 pgm_ref=>\@pgm_pssmlist, lib_ref => \@blp_list,
	 n_q => 1, msa_q => 1,
	 run_bkgd => 1,
	 lib_env=> $FAST_GNMS, remote=>1, have_ssr =>1,
	 pgm_args => "-q -w 80 -m 6 -b 500 -d 250",
	 use_query1 => 1, query2_type => 'lib',
	 get_lib_sub => \&get_lib,
	 link_url_ref =>"",
	 opts => {
	     %fa_opt_params,
	     eval => ["-E",\&get_safe_number],
	     etop => ["-F",\&get_safe_number],
	     zstat => ["-z",\&get_safe_number],
	 },
     },

     'hmm_search' => {
	 pgm_ref=> \@pgm_hmmlist,
	 lib_env => $FAST_LIBS,
	 n_q => 1,
	 run_bkgd => 1,
	 no_html => 1,
	 q_arg => 'hmm_query',
	 qt_arg => 'hmm_qtype',
	 qf_arg => 'hmm_query_file',
	 query1_type => 'tmp',
	 query2_type => 'tmp|lib', 
	 get_lib_sub => \&get_lib_full,
	 opts => {
	     eval => ["-E",\&get_safe_number],
	 },
	 footer => $hmm_footer,
     },

 );

# %page_links and %page_link_list are used for the list of links
# at the top of each page.

%page_links = 
  (
   select => { desc => qq(Search Databases with FASTA),
	       link => qq(fasta_www.cgi?rm=select),
	     },
   selectg => { desc => qq(Search Proteomes/Genomes),
		link => qq(fasta_www.cgi?rm=selectg)},
   blast => { desc => qq(Search Databases with BLAST),
	      link => qq(fasta_www.cgi?rm=blast)},
   mselect => { desc => qq(Search Databases with PSSMs),
		 link => qq(fasta_www.cgi?rm=mselect)},
   hmm_select => { desc => qq(Search Databases with HMMs),
		 link => qq(fasta_www.cgi?rm=hmm_select)},
   hmm_select2 => { desc => qq(Scan Sequences with HMMs),
		 link => qq(fasta_www.cgi?rm=hmm_select2)},
   compare => { desc => qq(Compare Two Sequences ),
		link => qq(fasta_www.cgi?rm=compare)
	      },
   shuffle => { desc => "Statistical Significance from Shuffles",
		link => "fasta_www.cgi?rm=shuffle"},
   lalign => { desc => "Find Internal Duplications (<b>lalign/plalign</b>)",
	       link => "fasta_www.cgi?rm=lalign"},
   misc1 =>  { desc => "Hydropathy/Secondary-Structure/<tt>seg</tt>",
	       link => "fasta_www.cgi?rm=misc1"}
  );

%page_link_list = 
  (
   select => [[qw(this selectg shuffle lalign misc1)],$select_text],
   selectg =>[[qw(this select shuffle lalign misc1)],$selectg_text],
   blast =>[[qw(this select shuffle lalign)], $blast_text],
   mselect =>[[qw(this blast select shuffle lalign)], $psiblast_text],
   hmm_select =>[[qw(this hmm_select2 select shuffle)], $hmm_select_text],
   hmm_select2 =>[[qw(this hmm_select select shuffle) ], $hmm_compare_text],
   shuffle =>[[qw(this select lalign)], $shuff_text],
   lalign =>[[qw(this select shuffle)], $lalign_text],
   compare =>[[qw(this select shuffle lalign)], $comp_text],
   misc1 =>[[qw(this select shuffle lalign)], $misc_text],
  );

sub query_from_mquery {
    my ($mquery) = @_;

    my ($title) = ($mquery =~ m/^(\S+)\s/);
    $title = "query" unless ($title);

    my $query = ">$title\n" . 
	join("", 
	     map { s/\W//g; "$_\n" } 
	     $mquery =~ m/(?:\A|\n\s*\n)\S+\s*([^\n]+)/sg #A. Mackey's secret code
	     );
}

sub clean_mquery {
    my ($mquery) = @_;

    unless ($mquery) { return "";}
    
    $mquery =~ s/\r\n/\n/sg;	# strip \r\n from Windows
    $mquery =~ s/\A(\s*\n)*//s;	# remove blank lines??
    $mquery =~ s/(\s*\n)*\Z//s;	# 
    $mquery =~ s/^CLUSTAL .*$//m;
    $mquery =~ s/^MUSCLE .*$//m;
    return $mquery;
}

sub get_bl_filter {
    my ($opt, $p_arg) = @_;
    if ($p_arg) {return $opt;}
    else {return "";}
}

%pgm_dev = ( pkd => {dev => "ps"},
	     pal => {dev => "SVG",
		     pgm_svg => "lav2svg -Z 1",
		     pgm_ps => "lav2ps -Z 1"},
	     pald => {dev => "SVG",
		      pgm_svg => "lav2svg -Z 1",
		      pgm_ps => "lav2ps -Z 1"},
	     );
1;
