
################################################################
# subroutines shared by FASTA_WWW.pm and CHAPS.pm to generate
# table-based pages
################################################################

use HTML::Entities;

# uppercase hash keys are (almost) always TMPL_VAR names
# find them and load them (TITLE, RUN_MODE, OPTION1, OPTION2, SUBMIT)
#
sub load_form {
  my ($form_name, $tmpl) = @_;

  for my $field ( keys %{$form_list{$form_name}}  ) {
    next unless ($field eq uc($field));
    if ($tmpl->query('name' => $field)) {
      $tmpl->param($field => $form_list{$form_name}->{$field});
    }
  }
}

################################################################
# get_option() and get_neg_option() are used as part of an argument
# parsing strategy that takes a hash_ref of the form:
# 	 opts => {
# 	     eval => ["-e", \&get_safe_number],
# 	     comp_stat => ["-C 1", \&get_option, "-C 0"],
# 	     seg_filter => ["-F T", \&get_option, "-F F"],
# 	     p_lib => ["-d", \&get_blib],
# 	     gap => ["-G", \&get_safe_number],
# 	     ext => ["-E", \&get_safe_number],
# 	     smatrix => ["", \&get_bmatrix],
# 	 },
#
# and checks each $q->param(key).  If found, then val[0] is added to
# the option string, based on the subsequent val[1]() code_ref. For If
# the code_ref is found, the function call is: code_ref->(val[0],
# $q->param($opt), $q, $opt); get_option() and get_net_option() only
# use val[0] and $q->param($opt) for options to be invoked, they must
# have name='option' value='1' or something
#
sub get_option {
  my ($opt, $p_arg) = @_;
  if ($p_arg) {
    return $opt;
  } else {
    return "";
  }
}

sub get_neg_option {
    my ($opt, $p_arg) = @_;
    if (!$p_arg) {return $opt;}
    else {return "";}
}

################
# get a safe number from $p_arg;
#
sub get_safe_number {
  my ($opt, $p_arg) = @_;

  unless ($p_arg) {return "";}

  if ($p_arg =~ m/DEFAULT/i) {return "";}

  ($p_arg) = ($p_arg =~ m/([E\d\-\.]+)/i);
  unless ($p_arg) {return "";}

  return "$opt $p_arg";
}

sub get_safe_string {
  my ($q,$p_arg) = @_;

  unless ($q->param($p_arg)) {return "";}

  my $value = $q->param($p_arg);

#  $value =~ s/[^$OK_CHARS]/_/go;
#  ($value) = ($value =~ m/([$OK_CHARS]+)/);

  return $value;
}

sub get_safe_filename {
  my ($q,$p_arg) = @_;

  unless ($q->param($p_arg)) {return "";}

  my $value=$q->param($p_arg);

  $value =~ s/[^$OK_CHARS]/_/go;

  ($value) = ($value =~ m/([$OK_CHARS]+)/);
  return $value;
}

sub get_safe_range {
  my ($opt, $p_arg) = @_;

  unless ($p_arg) {return "";}

  ($p_arg) = ($p_arg =~ m/(\d*\-?\d*)/);
  unless ($p_arg) {return "";}

  if ($opt) {return "$opt $p_arg";}
  else {return $p_arg;}
}

################
# get_file2file takes an uploaded file, copies it
# to a temporary file, and returns its
# uploaded name as an argument
#
# $opt -- command line option returned
# $p_arg  -- $q->param($arg)
# $q -- CGI query
# $arg -- parameter name
sub get_file2file {
  my ($opt, $p_arg, $q, $arg) = @_;

  # check to see if filename is available
  unless ($p_arg) {
    print STDERR "p_arg - missing: $opt\n";
    return "";
  }

  my $qfh = $q->upload($arg);

  if (!$qfh) {
    print STDERR "File upload failure: $p_arg\n";
    if ($q->cgi_error) {print STDERR $q->cgi_error();}
    return "";
  }

################
# capable of decoding PSSM:2
#
  my $buffer;
  my $data = '';
  while (read($qfh,$buffer,2048)) {$data .= $buffer;}
  close($qfh);

  if ($data =~ m/^.*\n*PSSM:2\n/s) {
      $data = bz2pssm($data);
  }

# create a temporary file
  my $tmp_fh = new File::Temp(DIR=>$TMP_DIR,
			      TEMPLATE => "FA_WWW_XXXXXX",
			      SUFFIX => ".asn1",
			      UNLINK => $DEF_UNLINK);
  chmod 0644, $tmp_fh->filename();

  print $tmp_fh $data;
  close($tmp_fh);
  push @tmp_fh_list, \$tmp_fh;

  return sprintf($opt, $tmp_fh->filename());
}

# modified 2/26/2007 to use HTML::Entities::encode to protect against XSS
#
sub fasta_error {

    my $msg = shift;

    return "<p /><hr /><p /><h2>". HTML::Entities::encode($msg) ." </h2><p /><hr /><p /></body></html>\n";
}

sub get_fastacmd {

  my ($db,$query) = @_;
  my $db_file;
  use vars qw($in $out $err);	# used for IPC::Run

  if (length($query) > 100) {
      return $query;
  }

  if ($db =~ m/P/i) {
      $db_file = "$BL_DB_DIR/nr";
      $db = "-pT";
  }
  else {
      # no nt fastacmd
      return "";
      $db_file = "$BL_DB_NT_DIR/nt";
      $db = "-pF";
  }

  $query =~ s/^\w+\|//;

  ($query) = ($query =~ m/(\w+)/);

  use vars qw($in $out $err);

  $in = "";

  my @cmd_list = split(/\s+/,"$BL_BIN_DIR/fastacmd -cT -d $db_file $db -tT -s $query");
#  print STDERR join(' ',@cmd_list),"\n";

  IPC::Run::run \@cmd_list, \$in, \$out, \$err; # timeout(240)

  if ($err) {
      print STDERR "fastacmd warning for query $query\n*** $err ***\n"; return "";
  }


  open(POS, '<', \$out) || return "";	# get output from fastacmd

  my $header = <POS>;
  my $line = "";

  return "" unless ($header);

  my @headers = split(/\001/,$header);
  
  for my $my_head ( @headers ) {
      if ($my_head =~ m/$query/i) {
	  $header = $my_head;
#	  print STDERR "$query:\n$header\n";
	  if ($header =~ m/\|sp\|/) {
	    $header =~ s/RecName: \w+=//g;
	    $header =~ s/AltName: \w+=//g;
	    $header =~ s/Short=/ /g;
	  }

	  last;
      }
  }

  while (<POS>) {
    $line=$line . $_;
  }
  close(POS);

  if ($header !~ m/^>/) {
    $header = '>' . $header;
  }

  return "$header\n$line";
}

sub get_ncbi {
    my ($db,$seq_in) = @_;
    use vars qw( $url $entry_line $parse );

    my $eutil_url = "http://eutils.ncbi.nih.gov/entrez/eutils/";

    $seq_in =~ s/gi\|//;

    if ($db =~ m/^P/i) {$db="Protein";}
    else {$db = "Nucleotide";}

#    print "<pre>seq_in: $seq_in\ndb: $db\n</pre>\n";

    if ($seq_in =~ m/^\d+$/) {
        $url= $eutil_url. "efetch.fcgi?db=$db&rettype=fasta&id=$seq_in";
        return get($url);
    }
    else {      # not gi, do a search first -
	$seq_in =~ s/\s+/+/g;
        $url="http://eutils.ncbi.nih.gov/entrez/eutils/esearch.fcgi?db=$db&usehistory=y&term=$seq_in";
        my $esearch_result = get($url);

#	print STDERR $esearch_result;

	my ($count, $querykey, $webenv) = ($esearch_result =~ 
	    m|<Count>(\d+)</Count>.*<QueryKey>(\d+)</QueryKey>.*<WebEnv>(\S+)</WebEnv>|s);

#	print STDERR "get_ncbi ::".$seq_in."::$count\n";
#	print STDERR "Count: $count - QueryKey: $querykey - WebEnv: $webenv\n";

	if ($count < 1) { return "";}

	my $efetch = "";
	my $retstart;
	my $retmax=100;

	for($retstart = 0; $retstart < $count; $retstart += $retmax) {
	    $url = $eutil_url . "efetch.fcgi?"
               . "rettype=fasta&retmode=text&"
	       . "retstart=$retstart&retmax=$retmax&"
               . "db=$db&query_key=$querykey&WebEnv=$webenv";
	
	    $efetch .= get($url);
	}
#	print STDERR $efetch;
	if ($efetch =~ m/\|sp\|/) {
	  $efetch =~ s/RecName: \w+=//g;
	  $efetch =~ s/AltName: \w+=//g;
	  $efetch =~ s/Short=/ /g;
	}

	return $efetch;
    }
}

# replaced 1-Nov-2006 with valid eutils calls
#
sub get_ncbi_old {

    my ($db,$seq_in) = @_;
    use vars qw( $url $entry_line $parse );

    $seq_in =~ s/gi\|//;

    if ($db =~ m/^P/i) {$db="Protein";}
    else {$db = "Nucleotide";}

#    print "<pre>seq_in: $seq_in\ndb: $db\n</pre>\n";


    if ($seq_in =~ m/^\d+$/) {
        $url="http://www.ncbi.nlm.nih.gov/entrez/query.fcgi?db=$db&cmd=Text&dopt=FASTA&uid=$seq_in";
        $entry_line = get $url;
        $parse = 0;
    }
    else {      # not gi, look for accession
        $url="http://www.ncbi.nlm.nih.gov/entrez/query.fcgi?db=$db&cmd=Search&doptcmdl=FASTA&term=$seq_in";
        $entry_line = get $url;
        $parse = 1;
    }

#    print $entry_line;
#    print "\n================\n";

    if (!$entry_line || $entry_line eq "" ||
        $entry_line =~/ No Documents Found /i ||
        $entry_line =~/temporarily unavailable/i) {

	$seq_in = "";
	return $seq_in;
    }

    if ($parse) {
        ($seq_in) = $entry_line =~ m/<pre>(.+?)<\/pre>/s;
        $seq_in =~ s/^.*>gi\|/>gi\|/;
    }
    else { $seq_in = $entry_line;}

    return $seq_in;
}

sub gp2fasta {
    my @entry = @_;

    my $locus = '';
    my $acc = '';
    my $gi = '';
    my $def = '';

    while ($_ = shift @entry) {
        if (!$locus && /^LOCUS/) {($locus) = /^LOCUS\s+(\w+)\s/;}
        elsif (!$def && /^DEFINITION/) {($def) = /^DEFINITION  (.+)$/;}
        elsif (!$acc && /^ACCESSION/) {($acc) = /^ACCESSION\s+(\w+)/;}
        elsif (!$gi && /^PID/) { ($gi) = /^PID\s+g(\d+)/; last;}
    };

    do {
        $_ = shift @entry;
    } until (/^ORIGIN/);

    my $seq = '';
    while ($_ = shift @entry) {
        if ($_ eq '//') {last;}
        $seq = $seq . substr($_,10,65);
    }
    $seq =~ s/\s+//g;
    $seq =~ s/(.{60})/$1\n/g;
    my $gp2fasta = ">gi|".$gi.'|db|'.$acc.'|'.$locus.' '.$def."\n";
    $gp2fasta = $gp2fasta . $seq;
    return $gp2fasta;
}

################################################################
# process a hash of input arguments that looks like:
# inputs => {
#   query => {SEARCH_QUERY => "this", SEARCH_FRM_ACC => "selected",
#             SEARCH_RANGE => &get_search_range},
#  query2 => {SEARCH_QUERY2 => "this", SEARCH_FRM_ACC2 => "selected"}
# }
################################################################
#
sub load_inputs {
  my ($form_href, $tmpl, $q) = @_;

# first deal with dependencies:
#
#   inp_dep_list => [qw(mquery query)]
# and:
#   inp_dep => {mquery => [mquery, \&clean_mquery],
#                query => [mquery, \&query_from_mquery]
#              }
# we look at the list of desired inputs and see if the inputs can be
# produced from other inputs
#
  for my $req_inp ( @{$form_href->{inp_dep_list}} ) {
  # can we resolve the dependency?
    if (exists $form_href->{inp_deps}->{$req_inp}) {
  # get the [name,\&function()] list
      my $req_inp_l = $form_href->{inp_deps}->{$req_inp};
  # if the names are the same, then do it, even if the argument is defined.
      if ($req_inp eq $req_inp_l->[0] )	{
	if (defined $q->param($req_inp_l->[0])) {
	  $q->param($req_inp,$req_inp_l->[1]($q->param($req_inp_l->[0])));
	}
      }
  # otherwise, only do it if the parameter has not been defined
      else {
	if (! defined($q->param($req_inp)) &&
	    defined $q->param($req_inp_l->[0])) {
	  $q->param($req_inp, $req_inp_l->[1]($q->param($req_inp_l->[0])));
	}
      }
    }
  }

#      "inputs" sets template vars in form from input parameters
#      inputs => { arg1 => {TMP_VAR1 => value, TMP_VAR2 => \&value_func, ...},
#
  my $input_href = $form_href->{inputs};

#    keys %$input_href are often "query", "query2", but also include
#    "remote", "msa_query", "hmm_query"
#
  for my $query ( keys %{$input_href} ) {   # query value from $q->param()
    if ($q->param($query)) {
	load_vars($input_href->{$query}, $tmpl, $q, $query);
    }
  }

}


################################################################
# load_vars() is called in two places, once as
#
# load_vars($form_href->{outputs}, $tmpl);
# where $href->{outputs} looks like:
#     { SUBMIT => 'Compare Sequences',
#      Q2_FILE_UP => 1,
#      SSR_FLAG => '1',
#      ...
#
# and once as:
#  for my $query ( keys %{$input_href} ) {   # query value from $q->param()
#    if ($q->param($query)) {load_vars($input_href->{$query}, $tmpl, $q, $query);}
#  }
#  where %$input_href looks like:
#     inputs =>
#    { query => {SEARCH_QUERY => "this", SEARCH_FRM_ACC => "selected",
#		SSR => \&get_query_range},
#      query2 => {SEARCH_QUERY2 => "this", SEARCH_FRM_ACC2 => "selected"}
#    },
#
################################################################
# In the former case (2 args), we just need to map TMPL_VARs to text
# strings. In the latter case (4 args), we are getting information
# from CGI parameters ($q) and possibly validating input or doing
# other transformations before setting the TMPL_VAR
#
################################################################
# take a hash_ref of the form:
#  outputs =>
#   { TITLE => qq(Misc. Protein Analysis)}
#  and fill in the TMPL_VAR (key, TITLE) with the value
#

sub load_vars {
  my ($input_href, $tmpl, $q, $query) = @_;

  for my $tmpl_var ( keys %{$input_href} ) { # keys are TMPL_VAR's

    my $value = $input_href->{$tmpl_var};    # get the TMPL_VAR name from the list

    if ($tmpl->query('name' => $tmpl_var)) { # check that it's in the template

      # if the value of TMPL_VAR is a function, execute it and set the variable
      if (ref($value) && ref($value) eq 'CODE' ) {
	$tmpl->param($tmpl_var => $value->($q,lc($tmpl_var)));
      }

      # otherwise if it is 'this', set it safely from an input parameter
      elsif ($value eq 'this' && $query && ref($q) eq 'CGI') {
	# some protection from XSS
	my $u_query = $q->param($query);
	$tmpl->param($tmpl_var => HTML::Entities::encode($u_query));
      }

      # otherwise (simplest case) just assign the $value to TMPL_VAR
      else {
	$tmpl->param($tmpl_var => $value);
      }
    }
  }
}

sub get_node_host {
  my $n_host = "";
  my @node_hosts = @NODE_HOSTS;
  my $node_ext = $NODE_EXT;

  if (-e $NODE_STATUS_FILE) {
    my @node_list = ();
    open(FH, $NODE_STATUS_FILE) || goto random;
    while (my $node_line = <FH>) {
      chomp $node_line;
      my %node_entry;
      @node_entry{('host', 'load')}= split(/:/,$node_line);
      push @node_list, \%node_entry;
    }
    close(FH);
    goto random unless (@node_list);
    @node_list = sort { $a->{load} <=> $b->{load} } @node_list;
    @node_hosts = map { $_->{host} } @node_list[0 .. scalar(@node_list)+1/2];
    $node_ext = "";
  }
# get it randomly
  random:
    $n_host = $node_hosts[int rand scalar @node_hosts];
# fill out full ip address
    $n_host .= $node_ext;

  return $n_host;
}

1;
